local CombatAlert = Rock:NewAddon("CombatAlert", "LibRockDB-1.0", "LibRockHook-1.0", "LibRockEvent-1.0", "LibRockTimer-1.0", "LibRockConfig-1.0", "LibRockConsole-1.0", "LibFuBarPlugin-3.0", "LibParser-4.0")
local self = CombatAlert

local _G = _G
local bit = _G.bit
local time = _G.time
local math = _G.math
local string = _G.string
local ipairs = _G.ipairs
local select = _G.select
local tonumber = _G.tonumber
local UIParent = _G.UIParent
local CreateFrame = _G.CreateFrame
local GetBuildInfo = _G.GetBuildInfo
local PlaySoundFile = _G.PlaySoundFile
local IsAddOnLoaded = _G.IsAddOnLoaded
local CombatText_AddMessage = _G.CombatText_AddMessage
local GetInventorySlotInfo = _G.GetInventorySlotInfo
local GetInventoryItemLink = _G.GetInventoryItemLink
local GetInventoryItemCooldown = _G.GetInventoryItemCooldown
local GetContainerNumSlots = _G.GetContainerNumSlots
local GetContainerItemLink = _G.GetContainerItemLink
local GetContainerItemCooldown = _G.GetContainerItemCooldown
local GetSpellName = _G.GetSpellName
local GetSpellCooldown = _G.GetSpellCooldown
local GetRealZoneText = _G.GetRealZoneText
local GetGameTime = _G.GetGameTime
local GetTime = _G.GetTime
local GetCVar = _G.GetCVar
local GetUnitInfoName = _G.GetUnitInfoName
local UnitName = _G.UnitName
local UnitFactionGroup = _G.UnitFactionGroup
local UnitCreatureType = _G.UnitCreatureType
local UnitCreatureFamily = _G.UnitCreatureFamily
local BOOKTYPE_SPELL = _G.BOOKTYPE_SPELL
local BOOKTYPE_PET = _G.BOOKTYPE_PET

self:AddSlashCommand('OnCommand', '/CombatAlert', '/cal', '/ca')
self:SetDatabase('CombatAlertDB', 'CombatAlertPCDB')
self:SetFuBarOption('iconPath', "Interface\\AddOns\\CombatAlert\\icon")
self:SetFuBarOption('hasNoColor', true)

local MESSAGE_INFO = 1
local MESSAGE_WARN = 2
local MESSAGE_ALERT = 3
local COOLDOWN_MINIMUM = 3
local COOLDOWN_MAXIMUM = 300
local COOLDOWN_MODIFIERS = { -- TODO: add modifier check
	["Cold Snap"] = {
		[3] = "Frost Nova",
		[1] = "Ice Block",
		[2] = "Frost Ward",
		[4] = "Ice Barrier",
		[5] = "Cone of Cold",
	},
	["Preparation"] = {
		[1] = "Evasion",
		[2] = "Sprint",
		[3] = "Vanish",
		[4] = "Cold Blood",
		[5] = "Adrenaline Rush",
		[6] = "Premeditation",
	},
	["Renataki's Charm of Beasts"] = {
		[1] = "Arcane Shot",
		[2] = "Aimed Shot",
		[3] = "Multi-Shot",
		[4] = "Volley",
	}
}
local COOLDOWN_GROUPS = { -- TODO: add group check
	["Arcane/Aimed Shots"] = {
		[1] = "Arcane Shot",
		[2] = "Aimed Shot",
	},
	["Reta/SW/Reck"] = {
		[1] = "Retaliation",
		[2] = "Shield Wall",
		[3] = "Recklessness",
	},
	["Shocks"] = {
		[1] = "Earth Shock",
		[2] = "Flame Shock",
		[3] = "Frost Shock",
	},
	["Traps"] = {
		[1] = "Freezing Trap",
		[2] = "Immolation Trap",
		[3] = "Explosive Trap",
		[4] = "Frost Trap",
	},
	["Wards"] = {
		[1] = "Frost Ward",
		[2] = "Fire Ward",
	}
}

function CombatAlert:OnInitialize()
	self.callbacks = {
		["OnFeedbackFrameMoved"] = function(caller, event, xoffset, yoffset)
			self.db.profile.x, self.db.profile.y = xoffset, yoffset
		end
	}
	self.feedback = LibStub("LibFeedback-0.1")
	self.sharedmedia = LibStub("LibSharedMedia-3.0")
	self.sharedmedia:Register('sound', "CA: Info", "Sound\\Doodad\\BellTollTribal.wav")
	self.sharedmedia:Register('sound', "CA: Warn", "Sound\\Doodad\\BellTollNightElf.wav")
	if UnitFactionGroup('player') == "Horde" then
		self.sharedmedia:Register('sound', "CA: Alert", "Sound\\Doodad\\BellTollHorde.wav")
	else
		self.sharedmedia:Register('sound', "CA: Alert", "Sound\\Doodad\\BellTollAlliance.wav")
	end
	self:SetDatabaseDefaults('profile', {
		x = 0,
		y = 150,
		lock = true,
		preset = false,
		damage = false,
		unknown = false,
		prediction = false,
		compact = false,
		collect = false,
		silent = false,
		infosound = "None",
		warnsound = "CA: Warn",
		alertsound = "CA: Alert",
		frame = "feedback",
		fonttype = "Arial Narrow",
		fontsize = "16",
		fontoutline = "",
		damagecolor = { r=0.0, g=1.0, b=0.0, a=1.0 },
		statecolor = { r=0.0, g=1.0, b=0.0, a=1.0 },
		cooldowncolor = { r=1.0, g=1.0, b=0.0, a=1.0 },
		debuffcolor = { r=0.5, g=0.0, b=1.0, a=1.0 },
		buffcolor = { r=0.5, g=0.0, b=0.5, a=1.0 },
		castcolor = { r=1.0, g=1.0, b=1.0, a=1.0 },
		fadecolor = { r=1.0, g=0.5, b=0.0, a=1.0 },
		dispelcolor = { r=1.0, g=0.0, b=0.0, a=1.0 },
		showIcon = true,
		showText = true,
	})
	self:SetDatabaseDefaults('char', {
		known = {},
		show = {},
		warn = {},
		alert = {},
	})
	self:SetConfigTable({
		type = 'group',
		name = "Combat Alert",
		desc = "Combat event announcer",
		icon = "Interface\\AddOns\\CombatAlert\\icon",
		args = {
			preset = {
				order = 100,
				type = 'boolean',
				name = "Preset mode",
				desc = "Enable preset event alerts",
				get = function(preset)
					return self.db.profile.preset
				end,
				set = function(preset)
					self.db.profile.preset = preset
				end,
			},
			damage = {
				order = 101,
				type = 'boolean',
				name = "Show damage",
				desc = "Show incoming damage",
				get = function(damage)
					return self.db.profile.damage
				end,
				set = function(damage)
					self.db.profile.damage = damage
					if self.db.profile.damage then
						if not self:HasHook("CombatText_AddMessage") then self:AddHook("CombatText_AddMessage") end
					else
						if self:HasHook("CombatText_AddMessage") then self:RemoveHook("CombatText_AddMessage") end
					end
				end,
				hidden = function()
					return self.db.profile.preset
				end,
			},
			unknown = {
				order = 102,
				type = 'boolean',
				name = "Allow unknown",
				desc = "Enable unknown unit events",
				get = function()
					return self.db.profile.unknown
				end,
				set = function(unknown)
					self.db.profile.unknown = unknown
				end,
				hidden = function()
					return self.db.profile.preset
				end,
			},
			prediction = {
				order = 103,
				type = 'boolean',
				name = "Predict victims",
				desc = "Enable cast victim prediction",
				get = function()
					return self.db.profile.prediction
				end,
				set = function(prediction)
					self.db.profile.prediction = prediction
				end,
				hidden = function()
					return self.db.profile.preset
				end,
			},
			compact = {
				order = 104,
				type = 'boolean',
				name = "Compact messages",
				desc = "Enable displaying compact alerts",
				get = function()
					return self.db.profile.compact
				end,
				set = function(compact)
					self.db.profile.compact = compact
				end,
			},
			collect = {
				order = 105,
				type = 'boolean',
				name = "Collect alerts",
				desc = "Enable new alert collection",
				get = function()
					return self.db.profile.collect
				end,
				set = function(collect)
					self.db.profile.collect = collect
				end,
			},
			silent = {
				order = 106,
				type = 'boolean',
				name = "Silent collect",
				desc = "Disable printing new alerts",
				get = function()
					return self.db.profile.silent
				end,
				set = function(silent)
					self.db.profile.silent = silent
				end,
				hidden = function()
					return not self.db.profile.collect
				end,
			},
			infosound = {
				order = 107,
				type = 'choice',
				name = "Info sound",
				desc = "Changes the info sound",
				get = function()
					return self.db.profile.infosound
				end,
				set = function(infosound)
					self.db.profile.infosound = infosound
					if self.db.profile.infosound == "None" then
						self:Print("info sound disabled!")
					else
						PlaySoundFile(self.sharedmedia:Fetch('sound', infosound))
					end
				end,
				choices = self.sharedmedia:List('sound'),
			},
			warnsound = {
				order = 108,
				type = 'choice',
				name = "Warn sound",
				desc = "Changes the warn sound",
				get = function()
					return self.db.profile.warnsound
				end,
				set = function(warnsound)
					self.db.profile.warnsound = warnsound
					if self.db.profile.warnsound == "None" then
						self:Print("warn sound disabled!")
					else
						PlaySoundFile(self.sharedmedia:Fetch('sound', warnsound))
					end
				end,
				choices = self.sharedmedia:List('sound'),
			},
			alertsound = {
				order = 109,
				type = 'choice',
				name = "Alert sound",
				desc = "Changes the alert sound",
				get = function()
					return self.db.profile.alertsound
				end,
				set = function(alertsound)
					self.db.profile.alertsound = alertsound
					if self.db.profile.alertsound == "None" then
						self:Print("alert sound disabled!")
					else
						PlaySoundFile(self.sharedmedia:Fetch('sound', alertsound))
					end
				end,
				choices = self.sharedmedia:List('sound'),
			},
			frame = {
				order = 110,
				type = 'choice',
				name = "Message frame",
				desc = "Choose the message frame to use",
				get = function()
					return self.db.profile.frame
				end,
				set = function(frame)
					if (frame == "sct" and IsAddOnLoaded("sct")) or (frame == "msbt" and IsAddOnLoaded("MikScrollingBattleText")) or (frame == "parrot" and IsAddOnLoaded("Parrot")) or (frame == "blizzard" and SHOW_COMBAT_TEXT == "1") or frame == "feedback" then
						self.db.profile.frame = frame
					else
						self:Print("frame isnt usable!")
					end
				end,
				choices = {
					["feedback"] = "Feedback",
					["blizzard"] = "Blizzard",
					["parrot"] = "Parrot",
					["msbt"] = "MSBT",
					["sct"] = "SCT",
				},
			},
			header1 = {
				order = 200,
				type = 'header',
				hidden = function()
					return self.db.profile.frame ~= "feedback"
				end,
			},
			lock = {
				order = 201,
				type = 'boolean',
				name = "Lock frame",
				desc = "Locks the message frame",
				get = function()
					return self.db.profile.lock
				end,
				set = function()
					self.db.profile.lock = not self.db.profile.lock
					if self.db.profile.lock then
						self.feedback:Lock("Output")
					else
						self.feedback:Unlock("Output")
					end
				end,
				hidden = function()
					return self.db.profile.frame ~= "feedback"
				end,
			},
			fonttype = {
				order = 202,
				type = 'choice',
				name = "Font type",
				desc = "Changes the message font type",
				get = function()
					return self.db.profile.fonttype
				end,
				set = function(fonttype)
					self.db.profile.fonttype = fonttype
					self:Announce("New font type selected")
				end,
				hidden = function()
					return self.db.profile.frame ~= "feedback"
				end,
				choices = self.sharedmedia:List('font'),
			},
			fontsize = {
				order = 203,
				type = 'choice',
				name = "Font size",
				desc = "Changes the message font size",
				get = function()
					return self.db.profile.fontsize
				end,
				set = function(fontsize)
					self.db.profile.fontsize = fontsize
					self:Announce("New font size selected")
				end,
				hidden = function()
					return self.db.profile.frame ~= "feedback"
				end,
				choices = {
					["16"] = "Tiny",
					["24"] = "Small",
					["32"] = "Medium",
					["48"] = "Large",
				},
			},
			fontoutline = {
				order = 204,
				type = 'choice',
				name = "Font outline",
				desc = "Changes the message font outline",
				get = function()
					return self.db.profile.fontoutline
				end,
				set = function(fontoutline)
					self.db.profile.fontoutline = fontoutline
					self:Announce("New font outline selected")
				end,
				hidden = function()
					return self.db.profile.frame ~= "feedback"
				end,
				choices = {
					[""] = "None",
					["OUTLINE"] = "Thin",
					["THICKOUTLINE"] = "Thick",
				},
			},
			header2 = {
				order = 300,
				type = 'header',
			},
			statecolor = {
				order = 301,
				type = 'color',
				name = "State color",
				desc = "Changes the state message color",
				get = function()
					return self.db.profile.statecolor.r, self.db.profile.statecolor.g, self.db.profile.statecolor.b, self.db.profile.statecolor.a
				end,
				set = function(red, green, blue, alpha)
					self.db.profile.statecolor = {
						r = red,
						g = green,
						b = blue,
						a = alpha or 1,
					}
				end,
			},
			cooldowncolor = {
				order = 302,
				type = 'color',
				name = "Cooldown color",
				desc = "Changes the cooldown message color",
				get = function()
					return self.db.profile.cooldowncolor.r, self.db.profile.cooldowncolor.g, self.db.profile.cooldowncolor.b, self.db.profile.cooldowncolor.a
				end,
				set = function(red, green, blue, alpha)
					self.db.profile.cooldowncolor = {
						r = red,
						g = green,
						b = blue,
						a = alpha or 1,
					}
				end,
			},
			debuffcolor = {
				order = 303,
				type = 'color',
				name = "Debuff color",
				desc = "Changes the debuff message color",
				get = function()
					return self.db.profile.debuffcolor.r, self.db.profile.debuffcolor.g, self.db.profile.debuffcolor.b, self.db.profile.debuffcolor.a
				end,
				set = function(red, green, blue, alpha)
					self.db.profile.debuffcolor = {
						r = red,
						g = green,
						b = blue,
						a = alpha or 1,
					}
				end,
			},
			buffcolor = {
				order = 304,
				type = 'color',
				name = "Buff color",
				desc = "Changes the buff message color",
				get = function()
					return self.db.profile.buffcolor.r, self.db.profile.buffcolor.g, self.db.profile.buffcolor.b, self.db.profile.buffcolor.a
				end,
				set = function(red, green, blue, alpha)
					self.db.profile.buffcolor = {
						r = red,
						g = green,
						b = blue,
						a = alpha or 1,
					}
				end,
			},
			castcolor = {
				order = 305,
				type = 'color',
				name = "Cast color",
				desc = "Changes the cast message color",
				get = function()
					return self.db.profile.castcolor.r, self.db.profile.castcolor.g, self.db.profile.castcolor.b, self.db.profile.castcolor.a
				end,
				set = function(red, green, blue, alpha)
					self.db.profile.castcolor = {
						r = red,
						g = green,
						b = blue,
						a = alpha or 1,
					}
				end,
			},
			fadecolor = {
				order = 306,
				type = 'color',
				name = "Fade color",
				desc = "Changes the fade message color",
				get = function()
					return self.db.profile.fadecolor.r, self.db.profile.fadecolor.g, self.db.profile.fadecolor.b, self.db.profile.fadecolor.a
				end,
				set = function(red, green, blue, alpha)
					self.db.profile.fadecolor = {
						r = red,
						g = green,
						b = blue,
						a = alpha or 1,
					}
				end,
			},
			dispelcolor = {
				order = 307,
				type = 'color',
				name = "Dispel color",
				desc = "Changes the dispel message color",
				get = function()
					return self.db.profile.dispelcolor.r, self.db.profile.dispelcolor.g, self.db.profile.dispelcolor.b, self.db.profile.dispelcolor.a
				end,
				set = function(red, green, blue, alpha)
					self.db.profile.dispelcolor = {
						r = red,
						g = green,
						b = blue,
						a = alpha or 1,
					}
				end,
			},
		},
	})
end

function CombatAlert:OnEnable()
	self.feedback:Create("Damage", { xoffset = self.db.profile.x, yoffset = self.db.profile.y, direction = "UP:ALT" })
	self.feedback:Create("Output", { xoffset = self.db.profile.x, yoffset = self.db.profile.y })
	self:SetVariables()
	self:SetListeners()
end

function CombatAlert:OnDisable() 
	self:RemoveAllHooks()
	self:RemoveAllTimers()
	self:RemoveAllEventListeners()
	self:RemoveAllParserListeners()
end

function CombatAlert:OnCommand(command, text)
	if text and string.find(text, "list%s(.+)") then
		local _, _, pattern = string.find(text, "list%s(.+)")
		self:List(pattern)
	elseif text and string.find(text, "insert%s(.+)") then
		local _, _, pattern = string.find(text, "insert%s(.+)")
		self:Insert(pattern)
	elseif text and string.find(text, "delete%s(.+)") then
		local _, _, pattern = string.find(text, "delete%s(.+)")
		self:Delete(pattern)
	elseif text and string.find(text, "search%s(.+)") then
		local _, _, pattern = string.find(text, "search%s(.+)")
		if self:Search(pattern) then
			self:Print("found")
		else
			self:Print("not found")
		end
	else
		self:Print("Invalid command, view the readme.txt file in this addons directory for valid commands")
	end
end

function CombatAlert:PLAYER_REGEN_ENABLED(caller, event)
	local eventType, combatState = "state", "nocombat"
	if self.db.profile.collect then
		if not self:Search(string.format("known %s %s", eventType, combatState)) then
			self:Insert(string.format("known %s %s", eventType, combatState))
			if not self.db.profile.silent then
				self:Print(string.format("New %s found: %s", eventType, combatState))
			end
		end
	end
	if self:Search(string.format("alert %s %s", eventType, combatState)) then
		if self.db.profile.compact then
			self:Announce(string.format("[-combat]"), self.db.profile.statecolor, MESSAGE_ALERT)
		else
			self:Announce(string.format("You have exited combat"), self.db.profile.statecolor, MESSAGE_ALERT)
		end
	elseif self:Search(string.format("warn %s %s", eventType, combatState)) then
		if self.db.profile.compact then
			self:Announce(string.format("[-combat]"), self.db.profile.statecolor, MESSAGE_WARN)
		else
			self:Announce(string.format("You have exited combat"), self.db.profile.statecolor, MESSAGE_WARN)
		end
	elseif self.db.profile.preset or self:Search(string.format("show %s %s", eventType, combatState)) then
		if self.db.profile.compact then
			self:Announce(string.format("[-combat]"), self.db.profile.statecolor, MESSAGE_INFO)
		else
			self:Announce(string.format("You have exited combat"), self.db.profile.statecolor, MESSAGE_INFO)
		end
	end
end

function CombatAlert:PLAYER_REGEN_DISABLED(caller, event)
	local eventType, combatState = "state", "incombat"
	if self.db.profile.collect then
		if not self:Search(string.format("known %s %s", eventType, combatState)) then
			self:Insert(string.format("known %s %s", eventType, combatState))
			if not self.db.profile.silent then
				self:Print(string.format("New %s found: %s", eventType, combatState))
			end
		end
	end
	if self:Search(string.format("alert %s %s", eventType, combatState)) then
		if self.db.profile.compact then
			self:Announce(string.format("[+combat]"), self.db.profile.statecolor, MESSAGE_ALERT)
		else
			self:Announce(string.format("You have entered combat"), self.db.profile.statecolor, MESSAGE_ALERT)
		end
	elseif self:Search(string.format("warn %s %s", eventType, combatState)) then
		if self.db.profile.compact then
			self:Announce(string.format("[+combat]"), self.db.profile.statecolor, MESSAGE_WARN)
		else
			self:Announce(string.format("You have entered combat"), self.db.profile.statecolor, MESSAGE_WARN)
		end
	elseif self.db.profile.preset or self:Search(string.format("show %s %s", eventType, combatState)) then
		if self.db.profile.compact then
			self:Announce(string.format("[+combat]"), self.db.profile.statecolor, MESSAGE_INFO)
		else
			self:Announce(string.format("You have entered combat"), self.db.profile.statecolor, MESSAGE_INFO)
		end
	end
end

function CombatAlert:BAG_UPDATE_COOLDOWN(caller, event)
	local slotIDs = {
		"HeadSlot",
		"NeckSlot",
		"ShoulderSlot",
		"BackSlot",
		"ChestSlot",
		"WristSlot",
		"HandsSlot",
		"WaistSlot",
		"LegsSlot",
		"FeetSlot",
		"Finger0Slot",
		"Finger1Slot",
		"Trinket0Slot",
		"Trinket1Slot",
		"MainHandSlot",
		"SecondaryHandSlot",
	}
	for _, slot in ipairs(slotIDs) do
		local slotID = GetInventorySlotInfo(slot)
		if GetInventoryItemLink('player', slotID) then
			local _, _, itemName = string.find(GetInventoryItemLink('player', slotID), "%[(.+)%]")
			local start, duration, hasCooldown = GetInventoryItemCooldown('player', slotID)
			if hasCooldown and start > 0 and duration > COOLDOWN_MINIMUM then
				self:AddTimer("COOLDOWN:" .. itemName, duration - (GetTime() - start), "OnCooldownEvent", itemName)
			end
		end
	end
	for containerID = 0, 4 do
		for slotID = 1, GetContainerNumSlots(containerID) do
			if GetContainerItemLink(containerID, slotID) then
				local _, _, itemName = string.find(GetContainerItemLink(containerID, slotID), "%[(.+)%]")
				local start, duration, hasCooldown = GetContainerItemCooldown(containerID, slotID)
				if hasCooldown and start > 0 and duration > COOLDOWN_MINIMUM then
					self:AddTimer("COOLDOWN:" .. itemName, duration - (GetTime() - start), "OnCooldownEvent", itemName)
				end
			end
		end
	end
end

function CombatAlert:SPELL_UPDATE_COOLDOWN(caller, event)
	local bookTypes = {
		['spell'] = BOOKTYPE_SPELL,
		['pet'] = BOOKTYPE_PET,
	}
	for bookType, bookID in pairs(bookTypes) do
		local spellID = 1
		local spellName = GetSpellName(spellID, bookID)
		while spellName do
			local start, duration, hasCooldown = GetSpellCooldown(spellID, bookID)
			if hasCooldown and start > 0 and duration > COOLDOWN_MINIMUM then
				self:AddTimer("COOLDOWN:" .. spellName, duration - (GetTime() - start), "OnCooldownEvent", spellName)
			end
			spellID = spellID + 1
			spellName = GetSpellName(spellID, bookID)
		end
	end
end

function CombatAlert:ZONE_CHANGED(caller, event)
end

function CombatAlert:PLAYER_ENTERING_WORLD(caller, event)
end

function CombatAlert:CombatLogEvent(caller, event, timestamp, eventtype, sourceGUID, sourceName, sourceFlags, destinationGUID, destinationName, destinationFlags, ...)
	local info = {}
	info.sourceName = sourceName
	if bit.band(sourceFlags, COMBATLOG_OBJECT_AFFILIATION_MINE) == COMBATLOG_OBJECT_AFFILIATION_MINE then
		if bit.band(sourceFlags, COMBATLOG_OBJECT_TYPE_PLAYER) == COMBATLOG_OBJECT_TYPE_PLAYER then
			info.sourceID = 'player'
		elseif bit.band(sourceFlags, COMBATLOG_OBJECT_TYPE_PET) == COMBATLOG_OBJECT_TYPE_PET then
			info.sourceID = 'pet'
		end
	elseif bit.band(sourceFlags, COMBATLOG_OBJECT_FOCUS) == COMBATLOG_OBJECT_FOCUS then
		info.sourceID = 'focus'
	elseif bit.band(sourceFlags, COMBATLOG_OBJECT_TARGET) == COMBATLOG_OBJECT_TARGET then
		info.sourceID = 'target'
	end
	if bit.band(sourceFlags, COMBATLOG_OBJECT_CONTROL_PLAYER) == COMBATLOG_OBJECT_CONTROL_PLAYER then
		info.sourcePvP = true
	else
		info.sourcePvP = false
	end
	info.recipientName = destinationName
	if bit.band(destinationFlags, COMBATLOG_OBJECT_AFFILIATION_MINE) == COMBATLOG_OBJECT_AFFILIATION_MINE then
		if bit.band(destinationFlags, COMBATLOG_OBJECT_TYPE_PLAYER) == COMBATLOG_OBJECT_TYPE_PLAYER then
			info.recipientID = 'player'
		elseif bit.band(destinationFlags, COMBATLOG_OBJECT_TYPE_PET) == COMBATLOG_OBJECT_TYPE_PET then
			info.recipientID = 'pet'
		end
	elseif bit.band(destinationFlags, COMBATLOG_OBJECT_FOCUS) == COMBATLOG_OBJECT_FOCUS then
		info.recipientID = 'focus'
	elseif bit.band(destinationFlags, COMBATLOG_OBJECT_TARGET) == COMBATLOG_OBJECT_TARGET then
		info.recipientID = 'target'
	end
	if bit.band(destinationFlags, COMBATLOG_OBJECT_CONTROL_PLAYER) == COMBATLOG_OBJECT_CONTROL_PLAYER then
		info.recipientPvP = true
	else
		info.recipientPvP = false
	end
	if eventtype == "SPELL_CAST_START" then
		info.eventType = "Cast" -- spellId, spellName, spellSchool
		info.isBegin = true
		info.abilityName = select(2, ...)
		if info.sourceName and info.abilityName then
			self:OnCastEvent(info)
		end
	elseif eventtype == "SPELL_AURA_APPLIED" then
		info.eventType = "Aura" -- spellId, spellName, spellSchool, auraType
		info.abilityName = select(2, ...)
		if select(4, ...) == AURA_TYPE_BUFF then
			info.isBuff = true
		elseif select(4, ...) == AURA_TYPE_DEBUFF then
			info.isBuff = false
		end
		if info.recipientName and info.isBuff and info.abilityName then
			self:OnBuffEvent(info)
		elseif info.recipientName and not info.isBuff and info.abilityName then
			self:OnDebuffEvent(info)
		end
	elseif eventtype == "SPELL_AURA_REMOVED" then
		info.eventType = "Fade" -- spellId, spellName, spellSchool, auraType
		info.abilityName = select(2, ...)
		if info.recipientName and info.abilityName then
			self:OnFadeEvent(info)
		end
	elseif eventtype == "SPELL_AURA_DISPELLED" or eventtype == "SPELL_AURA_STOLEN" then
		info.eventType = "Dispel" -- spellId, spellName, spellSchool, extraSpellId, extraSpellName, extraSpellSchool
		info.recipientAbilityName = select(2, ...)
		if info.recipientName and info.recipientAbilityName then
			self:OnDispelEvent(info)
		end
	end
end

function CombatAlert:OnCooldownEvent(cooldownName)
	local eventType = "cooldown"
	if self.db.profile.collect then
		if not self:Search(string.format("known %s %s", eventType, cooldownName)) then
			self:Insert(string.format("known %s %s", eventType, cooldownName))
			if not self.db.profile.silent then
				self:Print(string.format("New %s found: %s", eventType, cooldownName))
			end
		end
	end
	if self:Search(string.format("alert %s %s", eventType, cooldownName)) then
		if self.db.profile.compact then
			self:Announce(string.format("[%s]", cooldownName), self.db.profile.cooldowncolor, MESSAGE_ALERT)
		else
			self:Announce(string.format("%s ready", cooldownName), self.db.profile.cooldowncolor, MESSAGE_ALERT)
		end
	elseif self:Search(string.format("warn %s %s", eventType, cooldownName)) then
		if self.db.profile.compact then
			self:Announce(string.format("[%s]", cooldownName), self.db.profile.cooldowncolor, MESSAGE_WARN)
		else
			self:Announce(string.format("%s ready", cooldownName), self.db.profile.cooldowncolor, MESSAGE_WARN)
		end
	elseif self.db.profile.preset or self:Search(string.format("show %s %s", eventType, cooldownName)) then
		if self.db.profile.compact then
			self:Announce(string.format("[%s]", cooldownName), self.db.profile.cooldowncolor, MESSAGE_INFO)
		else
			self:Announce(string.format("%s ready", cooldownName), self.db.profile.cooldowncolor, MESSAGE_INFO)
		end
	end
end

function CombatAlert:OnCastEvent(info)
	local eventType, sourceID, recipientID, recipientName = "cast"
	if info.sourceID then
		sourceID = string.gsub(info.sourceID, "%d", "")
	elseif info.sourceName then
		sourceID = self:GetUnitInfo(info.sourceName)
	else
		sourceID = 'unknown'
	end
	if (self.db.profile.preset or self.db.profile.prediction) and info.sourceName then
		self:UpdateUnitInfo(self, "UpdateRequest")
		self:UpdateUnitInfo(self, "UpdateRequest", info.sourceID or sourceID)
		recipientID, recipientName = self:GetUnitTargetInfo(info.sourceName)
		if recipientID == 'player' then recipientName = "you" end
		if recipientID == 'target' then recipientName = "your target" end
		if recipientID == 'focus' then recipientName = "your focus" end
		if recipientID == 'pet' then recipientName = "your pet" end
		recipientID = string.gsub(recipientID, "%d", "")
	else
		recipientID = 'unknown'
	end
	if (not self.db.profile.unknown and not self.db.profile.preset and (sourceID == 'unknown' or (self.db.profile.prediction and recipientID == 'unknown'))) or not info.sourceName or not recipientName or not info.abilityName then
		return
	end
	if self.db.profile.collect then
		if not self:Search(string.format("known %s %s %s %s", eventType, sourceID, recipientID, info.abilityName)) then
			self:Insert(string.format("known %s %s %s %s", eventType, sourceID, recipientID, info.abilityName))
			if not self.db.profile.silent then
				self:Print(string.format("New %s found: %s %s %s", eventType, sourceID, recipientID, info.abilityName))
			end
		end
	end
	if self:Search(string.format("alert %s %s %s %s", eventType, sourceID, recipientID, info.abilityName)) then
		if sourceID == 'focus' and recipientID == 'unknown' then
			self:Announce(string.format("--- %s casts %s ---", info.sourceName, info.abilityName), self.db.profile.castcolor, MESSAGE_ALERT)
		elseif sourceID == 'target' and recipientID == 'unknown' then
			self:Announce(string.format("*** %s casts %s ***", info.sourceName, info.abilityName), self.db.profile.castcolor, MESSAGE_ALERT)
		elseif sourceID == 'focus' and recipientID ~= 'unknown' then
			self:Announce(string.format("--- %s casts %s at %s ---", info.sourceName, info.abilityName, recipientName), self.db.profile.castcolor, MESSAGE_ALERT)
		elseif sourceID == 'target' and recipientID ~= 'unknown' then
			self:Announce(string.format("*** %s casts %s at %s ***", info.sourceName, info.abilityName, recipientName), self.db.profile.castcolor, MESSAGE_ALERT)
		elseif recipientID ~= 'unknown' then
			self:Announce(string.format("%s casts %s at %s", info.sourceName, info.abilityName, recipientName), self.db.profile.castcolor, MESSAGE_ALERT)
		else
			self:Announce(string.format("%s casts %s", info.sourceName, info.abilityName), self.db.profile.castcolor, MESSAGE_ALERT)
		end
	elseif  self:Search(string.format("warn %s %s %s %s", eventType, sourceID, recipientID, info.abilityName)) then
		if sourceID == 'focus' and recipientID == 'unknown' then
			self:Announce(string.format("--- %s casts %s ---", info.sourceName, info.abilityName), self.db.profile.castcolor, MESSAGE_WARN)
		elseif sourceID == 'target' and recipientID == 'unknown' then
			self:Announce(string.format("*** %s casts %s ***", info.sourceName, info.abilityName), self.db.profile.castcolor, MESSAGE_WARN)
		elseif sourceID == 'focus' and recipientID ~= 'unknown' then
			self:Announce(string.format("--- %s casts %s at %s ---", info.sourceName, info.abilityName, recipientName), self.db.profile.castcolor, MESSAGE_WARN)
		elseif sourceID == 'target' and recipientID ~= 'unknown' then
			self:Announce(string.format("*** %s casts %s at %s ***", info.sourceName, info.abilityName, recipientName), self.db.profile.castcolor, MESSAGE_WARN)
		elseif recipientID ~= 'unknown' then
			self:Announce(string.format("%s casts %s at %s", info.sourceName, info.abilityName, recipientName), self.db.profile.castcolor, MESSAGE_WARN)
		else
			self:Announce(string.format("%s casts %s", info.sourceName, info.abilityName), self.db.profile.castcolor, MESSAGE_WARN)
		end
	elseif (self.db.profile.preset and (sourceID == 'target' or sourceID == 'focus')) or self:Search(string.format("show %s %s %s %s", eventType, sourceID, recipientID, info.abilityName)) then
		if sourceID == 'focus' and recipientID == 'unknown' then
			self:Announce(string.format("--- %s casts %s ---", info.sourceName, info.abilityName), self.db.profile.castcolor, MESSAGE_SHOW)
		elseif sourceID == 'target' and recipientID == 'unknown' then
			self:Announce(string.format("*** %s casts %s ***", info.sourceName, info.abilityName), self.db.profile.castcolor, MESSAGE_SHOW)
		elseif sourceID == 'focus' and recipientID ~= 'unknown' then
			self:Announce(string.format("--- %s casts %s at %s ---", info.sourceName, info.abilityName, recipientName), self.db.profile.castcolor, MESSAGE_SHOW)
		elseif sourceID == 'target' and recipientID ~= 'unknown' then
			self:Announce(string.format("*** %s casts %s at %s ***", info.sourceName, info.abilityName, recipientName), self.db.profile.castcolor, MESSAGE_SHOW)
		elseif recipientID ~= 'unknown' then
			self:Announce(string.format("%s casts %s at %s", info.sourceName, info.abilityName, recipientName), self.db.profile.castcolor, MESSAGE_SHOW)
		else
			self:Announce(string.format("%s casts %s", info.sourceName, info.abilityName), self.db.profile.castcolor, MESSAGE_SHOW)
		end
	end
end

function CombatAlert:OnBuffEvent(info)
	local eventType, recipientID = "buff"
	if info.recipientID then
		recipientID = string.gsub(info.recipientID, "%d", "")
	elseif info.recipientName then
		recipientID = self:GetUnitInfo(info.recipientName)
	else
		recipientID = 'unknown'
	end
	if (not self.db.profile.unknown and not self.db.profile.preset and recipientID == 'unknown') or not info.recipientName or not info.abilityName then
		return
	end
	if self.db.profile.collect then
		if not self:Search(string.format("known %s %s %s", eventType, recipientID, info.abilityName)) then
			self:Insert(string.format("known %s %s %s", eventType, recipientID, info.abilityName))
			if not self.db.profile.silent then
				self:Print(string.format("New %s found: %s %s", eventType, recipientID, info.abilityName))
			end
		end
	end
	if self:Search(string.format("alert %s %s %s", eventType, recipientID, info.abilityName)) then
		if recipientID == 'player' then
			if self.db.profile.compact then
				self:Announce(string.format("[+%s]", info.abilityName), self.db.profile.buffcolor, MESSAGE_ALERT)
			else
				self:Announce(string.format("You gain %s", info.abilityName), self.db.profile.buffcolor, MESSAGE_ALERT)
			end
		elseif recipientID == 'pet' then
			self:Announce(string.format("Your pet gains %s", info.abilityName), self.db.profile.buffcolor, MESSAGE_ALERT)
		elseif recipientID == 'focus' then
			self:Announce(string.format("--- %s gains %s ---", info.recipientName, info.abilityName), self.db.profile.buffcolor, MESSAGE_ALERT)
		elseif recipientID == 'target' then
			self:Announce(string.format("*** %s gains %s ***", info.recipientName, info.abilityName), self.db.profile.buffcolor, MESSAGE_ALERT)
		else
			self:Announce(string.format("%s gains %s", info.recipientName, info.abilityName), self.db.profile.buffcolor, MESSAGE_ALERT)
		end
	elseif (self.db.profile.preset and (recipientID == 'target' or recipientID == 'focus')) or self:Search(string.format("warn %s %s %s", eventType, recipientID, info.abilityName)) then
		if recipientID == 'player' then
			if self.db.profile.compact then
				self:Announce(string.format("[+%s]", info.abilityName), self.db.profile.buffcolor, MESSAGE_WARN)
			else
				self:Announce(string.format("You gain %s", info.abilityName), self.db.profile.buffcolor, MESSAGE_WARN)
			end
		elseif recipientID == 'pet' then
			self:Announce(string.format("Your pet gains %s", info.abilityName), self.db.profile.buffcolor, MESSAGE_WARN)
		elseif recipientID == 'focus' then
			self:Announce(string.format("--- %s gains %s ---", info.recipientName, info.abilityName), self.db.profile.buffcolor, MESSAGE_WARN)
		elseif recipientID == 'target' then
			self:Announce(string.format("*** %s gains %s ***", info.recipientName, info.abilityName), self.db.profile.buffcolor, MESSAGE_WARN)
		else
			self:Announce(string.format("%s gains %s", info.recipientName, info.abilityName), self.db.profile.buffcolor, MESSAGE_WARN)
		end
	elseif (self.db.profile.preset and recipientID == 'player') or self:Search(string.format("show %s %s %s", eventType, recipientID, info.abilityName)) then
		if recipientID == 'player' then
			if self.db.profile.compact then
				self:Announce(string.format("[+%s]", info.abilityName), self.db.profile.buffcolor, MESSAGE_INFO)
			else
				self:Announce(string.format("You gain %s", info.abilityName), self.db.profile.buffcolor, MESSAGE_INFO)
			end
		elseif recipientID == 'pet' then
			self:Announce(string.format("Your pet gains %s", info.abilityName), self.db.profile.buffcolor, MESSAGE_INFO)
		elseif recipientID == 'focus' then
			self:Announce(string.format("--- %s gains %s ---", info.recipientName, info.abilityName), self.db.profile.buffcolor, MESSAGE_INFO)
		elseif recipientID == 'target' then
			self:Announce(string.format("*** %s gains %s ***", info.recipientName, info.abilityName), self.db.profile.buffcolor, MESSAGE_INFO)
		else
			self:Announce(string.format("%s gains %s", info.recipientName, info.abilityName), self.db.profile.buffcolor, MESSAGE_INFO)
		end
	end
end

function CombatAlert:OnDebuffEvent(info)
	local eventType, recipientID = "debuff"
	if info.recipientID then
		recipientID = string.gsub(info.recipientID, "%d", "")
	elseif info.recipientName then
		recipientID = self:GetUnitInfo(info.recipientName)
	else
		recipientID = 'unknown'
	end
	if (not self.db.profile.unknown and not self.db.profile.preset and recipientID == 'unknown') or not info.recipientName or not info.abilityName then
		return
	end
	if self.db.profile.collect then
		if not self:Search(string.format("known %s %s %s", eventType, recipientID, info.abilityName)) then
			self:Insert(string.format("known %s %s %s", eventType, recipientID, info.abilityName))
			if not self.db.profile.silent then
				self:Print(string.format("New %s found: %s %s", eventType, recipientID, info.abilityName))
			end
		end
	end
	if self:Search(string.format("alert %s %s %s", eventType, recipientID, info.abilityName)) then
		if recipientID == 'player' then
			if self.db.profile.compact then
				self:Announce(string.format("[+%s]", info.abilityName), self.db.profile.debuffcolor, MESSAGE_ALERT)
			else
				self:Announce(string.format("You gain %s", info.abilityName), self.db.profile.debuffcolor, MESSAGE_ALERT)
			end
		elseif recipientID == 'pet' then
			self:Announce(string.format("Your pet gains %s", info.abilityName), self.db.profile.debuffcolor, MESSAGE_ALERT)
		elseif recipientID == 'focus' then
			self:Announce(string.format("--- %s gains %s ---", info.recipientName, info.abilityName), self.db.profile.debuffcolor, MESSAGE_ALERT)
		elseif recipientID == 'target' then
			self:Announce(string.format("*** %s gains %s ***", info.recipientName, info.abilityName), self.db.profile.debuffcolor, MESSAGE_ALERT)
		else
			self:Announce(string.format("%s gains %s", info.recipientName, info.abilityName), self.db.profile.debuffcolor, MESSAGE_ALERT)
		end
	elseif self:Search(string.format("warn %s %s %s", eventType, recipientID, info.abilityName)) then
		if recipientID == 'player' then
			if self.db.profile.compact then
				self:Announce(string.format("[+%s]", info.abilityName), self.db.profile.debuffcolor, MESSAGE_WARN)
			else
				self:Announce(string.format("You gain %s", info.abilityName), self.db.profile.debuffcolor, MESSAGE_WARN)
			end
		elseif recipientID == 'pet' then
			self:Announce(string.format("Your pet gains %s", info.abilityName), self.db.profile.debuffcolor, MESSAGE_WARN)
		elseif recipientID == 'focus' then
			self:Announce(string.format("--- %s gains %s ---", info.recipientName, info.abilityName), self.db.profile.debuffcolor, MESSAGE_WARN)
		elseif recipientID == 'target' then
			self:Announce(string.format("*** %s gains %s ***", info.recipientName, info.abilityName), self.db.profile.debuffcolor, MESSAGE_WARN)
		else
			self:Announce(string.format("%s gains %s", info.recipientName, info.abilityName), self.db.profile.debuffcolor, MESSAGE_WARN)
		end
	elseif (self.db.profile.preset and recipientID == 'player') or self:Search(string.format("show %s %s %s", eventType, recipientID, info.abilityName)) then
		if recipientID == 'player' then
			if self.db.profile.compact then
				self:Announce(string.format("[+%s]", info.abilityName), self.db.profile.debuffcolor, MESSAGE_INFO)
			else
				self:Announce(string.format("You gain %s", info.abilityName), self.db.profile.debuffcolor, MESSAGE_INFO)
			end
		elseif recipientID == 'pet' then
			self:Announce(string.format("Your pet gains %s", info.abilityName), self.db.profile.debuffcolor, MESSAGE_INFO)
		elseif recipientID == 'focus' then
			self:Announce(string.format("--- %s gains %s ---", info.recipientName, info.abilityName), self.db.profile.debuffcolor, MESSAGE_INFO)
		elseif recipientID == 'target' then
			self:Announce(string.format("*** %s gains %s ***", info.recipientName, info.abilityName), self.db.profile.debuffcolor, MESSAGE_INFO)
		else
			self:Announce(string.format("%s gains %s", info.recipientName, info.abilityName), self.db.profile.debuffcolor, MESSAGE_INFO)
		end
	end
end

function CombatAlert:OnDispelEvent(info)
	local eventType, recipientID = "dispel"
	if info.recipientID then
		recipientID = string.gsub(info.recipientID, "%d", "")
	elseif info.recipientName then
		recipientID = self:GetUnitInfo(info.recipientName)
	else
		recipientID = 'unknown'
	end
	if (not self.db.profile.unknown and not self.db.profile.preset and recipientID == 'unknown') or not info.recipientName or not info.recipientAbilityName then
		return
	end
	if self.db.profile.collect then
		if not self:Search(string.format("known %s %s %s", eventType, recipientID, info.recipientAbilityName)) then
			self:Insert(string.format("known %s %s %s", eventType, recipientID, info.recipientAbilityName))
			if not self.db.profile.silent then
				self:Print(string.format("New %s found: %s %s", eventType, recipientID, info.recipientAbilityName))
			end
		end
	end
	if self:Search(string.format("alert %s %s %s", eventType, recipientID, info.recipientAbilityName)) then
		if recipientID == 'player' then
			if self.db.profile.compact then
				self:Announce(string.format("[-%s]", info.recipientAbilityName), self.db.profile.dispelcolor, MESSAGE_ALERT)
			else
				self:Announce(string.format("%s is dispelled from you", info.recipientAbilityName), self.db.profile.dispelcolor, MESSAGE_ALERT)
			end
		elseif recipientID == 'pet' then
			self:Announce(string.format("%s is dispelled from your pet", info.recipientAbilityName), self.db.profile.dispelcolor, MESSAGE_ALERT)
		elseif recipientID == 'focus' then
			self:Announce(string.format("--- %s is dispelled from %s ---", info.recipientAbilityName, info.recipientName), self.db.profile.dispelcolor, MESSAGE_ALERT)
		elseif recipientID == 'target' then
			self:Announce(string.format("*** %s is dispelled from %s ***", info.recipientAbilityName, info.recipientName), self.db.profile.dispelcolor, MESSAGE_ALERT)
		else
			self:Announce(string.format("%s is dispelled from %s", info.recipientAbilityName, info.recipientName), self.db.profile.dispelcolor, MESSAGE_ALERT)
		end
	elseif (self.db.profile.preset and (recipientID == 'target' or recipientID == 'focus')) or self:Search(string.format("warn %s %s %s", eventType, recipientID, info.recipientAbilityName)) then
		if recipientID == 'player' then
			if self.db.profile.compact then
				self:Announce(string.format("[-%s]", info.recipientAbilityName), self.db.profile.dispelcolor, MESSAGE_WARN)
			else
				self:Announce(string.format("%s is dispelled from you", info.recipientAbilityName), self.db.profile.dispelcolor, MESSAGE_WARN)
			end
		elseif recipientID == 'pet' then
			self:Announce(string.format("%s is dispelled from your pet", info.recipientAbilityName), self.db.profile.dispelcolor, MESSAGE_WARN)
		elseif recipientID == 'focus' then
			self:Announce(string.format("--- %s is dispelled from %s ---", info.recipientAbilityName, info.recipientName), self.db.profile.dispelcolor, MESSAGE_WARN)
		elseif recipientID == 'target' then
			self:Announce(string.format("*** %s is dispelled from %s ***", info.recipientAbilityName, info.recipientName), self.db.profile.dispelcolor, MESSAGE_WARN)
		else
			self:Announce(string.format("%s is dispelled from %s", info.recipientAbilityName, info.recipientName), self.db.profile.dispelcolor, MESSAGE_WARN)
		end
	elseif (self.db.profile.preset and recipientID == 'player') or self:Search(string.format("show %s %s %s", eventType, recipientID, info.recipientAbilityName)) then
		if recipientID == 'player' then
			if self.db.profile.compact then
				self:Announce(string.format("[-%s]", info.recipientAbilityName), self.db.profile.dispelcolor, MESSAGE_INFO)
			else
				self:Announce(string.format("%s is dispelled from you", info.recipientAbilityName), self.db.profile.dispelcolor, MESSAGE_INFO)
			end
		elseif recipientID == 'pet' then
			self:Announce(string.format("%s is dispelled from your pet", info.recipientAbilityName), self.db.profile.dispelcolor, MESSAGE_INFO)
		elseif recipientID == 'focus' then
			self:Announce(string.format("--- %s is dispelled from %s ---", info.recipientAbilityName, info.recipientName), self.db.profile.dispelcolor, MESSAGE_INFO)
		elseif recipientID == 'target' then
			self:Announce(string.format("*** %s is dispelled from %s ***", info.recipientAbilityName, info.recipientName), self.db.profile.dispelcolor, MESSAGE_INFO)
		else
			self:Announce(string.format("%s is dispelled from %s", info.recipientAbilityName, info.recipientName), self.db.profile.dispelcolor, MESSAGE_INFO)
		end
	end
end

function CombatAlert:OnFadeEvent(info)
	local eventType, recipientID = "fade"
	if info.recipientID then
		recipientID = string.gsub(info.recipientID, "%d", "")
	elseif info.recipientName then
		recipientID = self:GetUnitInfo(info.recipientName)
	else
		recipientID = 'unknown'
	end
	if (not self.db.profile.unknown and not self.db.profile.preset and recipientID == 'unknown') or not info.recipientName or not info.abilityName then
		return
	end
	if self.db.profile.collect then
		if not self:Search(string.format("known %s %s %s", eventType, recipientID, info.abilityName)) then
			self:Insert(string.format("known %s %s %s", eventType, recipientID, info.abilityName))
			if not self.db.profile.silent then
				self:Print(string.format("New %s found: %s %s", eventType, recipientID, info.abilityName))
			end
		end
	end
	if self:Search(string.format("alert %s %s %s", eventType, recipientID, info.abilityName)) then
		if recipientID == 'player' then
			if self.db.profile.compact then
				self:Announce(string.format("[-%s]", info.abilityName), self.db.profile.fadecolor, MESSAGE_ALERT)
			else
				self:Announce(string.format("%s fades from you", info.abilityName), self.db.profile.fadecolor, MESSAGE_ALERT)
			end
		elseif recipientID == 'pet' then
			self:Announce(string.format("%s fades from your pet", info.abilityName), self.db.profile.fadecolor, MESSAGE_ALERT)
		elseif recipientID == 'focus' then
			self:Announce(string.format("--- %s fades from %s ---", info.abilityName, info.recipientName), self.db.profile.fadecolor, MESSAGE_ALERT)
		elseif recipientID == 'target' then
			self:Announce(string.format("*** %s fades from %s ***", info.abilityName, info.recipientName), self.db.profile.fadecolor, MESSAGE_ALERT)
		else
			self:Announce(string.format("%s fades from %s", info.abilityName, info.recipientName), self.db.profile.fadecolor, MESSAGE_ALERT)
		end
	elseif self:Search(string.format("warn %s %s %s", eventType, recipientID, info.abilityName)) then
		if recipientID == 'player' then
			if self.db.profile.compact then
				self:Announce(string.format("[-%s]", info.abilityName), self.db.profile.fadecolor, MESSAGE_WARN)
			else
				self:Announce(string.format("%s fades from you", info.abilityName), self.db.profile.fadecolor, MESSAGE_WARN)
			end
		elseif recipientID == 'pet' then
			self:Announce(string.format("%s fades from your pet", info.abilityName), self.db.profile.fadecolor, MESSAGE_WARN)
		elseif recipientID == 'focus' then
			self:Announce(string.format("--- %s fades from %s ---", info.abilityName, info.recipientName), self.db.profile.fadecolor, MESSAGE_WARN)
		elseif recipientID == 'target' then
			self:Announce(string.format("*** %s fades from %s ***", info.abilityName, info.recipientName), self.db.profile.fadecolor, MESSAGE_WARN)
		else
			self:Announce(string.format("%s fades from %s", info.abilityName, info.recipientName), self.db.profile.fadecolor, MESSAGE_WARN)
		end
	elseif (self.db.profile.preset and (recipientID == 'player' or recipientID == 'target' or recipientID == 'focus')) or self:Search(string.format("show %s %s %s", eventType, recipientID, info.abilityName)) then
		if recipientID == 'player' then
			if self.db.profile.compact then
				self:Announce(string.format("[-%s]", info.abilityName), self.db.profile.fadecolor, MESSAGE_INFO)
			else
				self:Announce(string.format("%s fades from you", info.abilityName), self.db.profile.fadecolor, MESSAGE_INFO)
			end
		elseif recipientID == 'pet' then
			self:Announce(string.format("%s fades from your pet", info.abilityName), self.db.profile.fadecolor, MESSAGE_INFO)
		elseif recipientID == 'focus' then
			self:Announce(string.format("--- %s fades from %s ---", info.abilityName, info.recipientName), self.db.profile.fadecolor, MESSAGE_INFO)
		elseif recipientID == 'target' then
			self:Announce(string.format("*** %s fades from %s ***", info.abilityName, info.recipientName), self.db.profile.fadecolor, MESSAGE_INFO)
		else
			self:Announce(string.format("%s fades from %s", info.abilityName, info.recipientName), self.db.profile.fadecolor, MESSAGE_INFO)
		end
	end
end

function CombatAlert:CombatText_AddMessage(message, ...)
	if type(message) == "string" then
		self.feedback:AddData("Damage", {
			text = message,
			sticky = true,
			fonttype = self.sharedmedia:Fetch('font', self.db.profile.fonttype),
			fontsize = tonumber(self.db.profile.fontsize),
			fontoutline = self.db.profile.fontoutline,
			fontcolor = self.db.profile.damagecolor
		})
	end
end

function CombatAlert:Announce(message, color, type)
	local messagecolor = color or { r = 1.0, g = 1.0, b = 1.0, a = 1.0 }
	if IsAddOnLoaded("sct") and self.db.profile.frame == "sct" then
		SCT:DisplayMessage(message, messagecolor)
	elseif IsAddOnLoaded("Parrot") and self.db.profile.frame == "parrot" then
		Parrot:ShowMessage(message, "Notification", false, messagecolor.r, messagecolor.g, messagecolor.b)
	elseif IsAddOnLoaded("MikScrollingBattleText") and self.db.profile.frame == "msbt" then
		MikSBT.DisplayMessage(message, MikSBT.DISPLAYTYPE_NOTIFICATION, false, messagecolor.r, messagecolor.g, messagecolor.b)
	elseif SHOW_COMBAT_TEXT == "1" and self.db.profile.frame == "blizzard" then
		CombatText_AddMessage(message, COMBAT_TEXT_SCROLL_FUNCTION, messagecolor.r, messagecolor.g, messagecolor.b)
	else
		self.feedback:AddData("Output", {
			text = message,
			sticky = type and type == MESSAGE_ALERT,
			fonttype = self.sharedmedia:Fetch('font', self.db.profile.fonttype),
			fontsize = tonumber(self.db.profile.fontsize),
			fontoutline = self.db.profile.fontoutline,
			fontcolor = messagecolor
		})
	end
	if type and type == MESSAGE_ALERT and self.db.profile.alertsound and self.db.profile.alertsound ~= "None" then
		PlaySoundFile(self.sharedmedia:Fetch('sound', self.db.profile.alertsound))
	elseif type and type == MESSAGE_WARN and self.db.profile.warnsound and self.db.profile.warnsound ~= "None" then
		PlaySoundFile(self.sharedmedia:Fetch('sound', self.db.profile.warnsound))
	elseif type and type == MESSAGE_INFO and self.db.profile.infosound and self.db.profile.infosound ~= "None" then
		PlaySoundFile(self.sharedmedia:Fetch('sound', self.db.profile.infosound))
	end
end

function CombatAlert:Insert(pattern)
	local _, _, type = string.find(pattern, "(%a+)")
	local _, _, value = string.find(pattern, "%a+%s(.+)")
	if type and self.db.char[type] and value then
		table.insert(self.db.char[type], value)
		table.sort(self.db.char[type])
	end
end

function CombatAlert:Search(pattern) -- TODO: re-check the way this works and maybe make a local search for self-use
	local _, _, type = string.find(pattern, "(%a+)")
	local _, _, value = string.find(pattern, "%a+%s(.+)")
	if type and self.db.char[type] and value then
		value = string.gsub(value, "*", ".+")
		for i, v in ipairs(self.db.char[type]) do
			v = string.gsub(v, "*", ".+")
			if string.find(v, value) or string.find(value, v) then
				return true
			end
		end
	end
	return false
end

function CombatAlert:Delete(pattern)
	local _, _, type = string.find(pattern, "(%a+)")
	local _, _, value = string.find(pattern, "%a+%s(.+)")
	if type and self.db.char[type] then
		if not value then
			self.db.char[type] = {}
		else
			value = string.gsub(value, "*", ".+")
			value = string.gsub(value, "\(.+\)", "")
			for i, v in ipairs(self.db.char[type]) do
				if string.find(v, value) then
					table.remove(self.db.char[type], i)
				end
			end
		end
	end
end

function CombatAlert:List(pattern)
	local _, _, type = string.find(pattern, "(%a+)")
	local _, _, value = string.find(pattern, "%a+%s(.+)")
	if type and self.db.char[type] then
		if value then
			value = string.gsub(value, "*", ".+")
		end
		for i, v in ipairs(self.db.char[type]) do
			if not value or string.find(v, value) then
				self:Print(string.format("%s %s", type, v))
			end
		end
	end
end

function CombatAlert:UpdateUnitInfo(caller, event, unit)
	if unit then
		if UnitExists(unit) then
			self:SetUnitInfo(caller, event, unit)
		else
			self.otherunits[unit] = nil
		end
	else
		for u in pairs(self.otherunits) do
			if UnitExists(u) then
				self:SetUnitInfo(caller, event, u)
			else
				self.otherunits[u] = nil
			end
		end
	end
end

function CombatAlert:SetUnitInfo(caller, event, unit)
	if not self.units[unit] and unit ~= 'npc' then
		local target, targetname = nil, UnitName(unit .. 'target')
		if targetname then
			for u in pairs(self.otherunits) do
				if self.otherunits[u].name == targetname then
					target = u
				end 
			end
			for u in pairs(self.units) do
				if UnitName(u) == targetname then
					target = u
				end 
			end
		end
		self.otherunits[unit] = {}
		self.otherunits[unit].name = UnitName(unit)
		self.otherunits[unit].target = target or 'unknown'
		self.otherunits[unit].targetname = targetname or "Unknown"
	end
end

function CombatAlert:GetUnitTargetInfo(unitname)
	for u in pairs(self.units) do
		local name = UnitName(u)
		if name and name == unitname then
			local target = function(u) if u == 'player' then return 'target' else return u .. 'target' end end
			local targetname = UnitName(target)
			if targetname and targetname == unitname then
				return target, targetname
			end
		end
	end
	for u in pairs(self.otherunits) do
		if self.otherunits[u].name == unitname then
			return self.otherunits[u].target, self.otherunits[u].targetname
		end
	end
	return 'unknown'
end

function CombatAlert:GetUnitInfo(unitname)
	for u in pairs(self.units) do
		local name = UnitName(u)
		if name and name == unitname then
			return u
		end
	end
	for u in pairs(self.otherunits) do
		if self.otherunits[u].name == unitname then
			return u
		end
	end
	for u in pairs(self.otherunits) do
		if self.otherunits[u].targetname == unitname then
			return self.otherunits[u].target
		end
	end
	return 'unknown'
end

function CombatAlert:SetVariables()
	self.location = GetRealZoneText() or ""
	self.name = UnitName('player')
	self.units, self.otherunits = { 'player', 'pet', 'focus', 'target', 'pettarget', 'mouseover' }, {}
end

function CombatAlert:SetListeners()
	local version, build, date = GetBuildInfo()
	self:RemoveAllEventListeners()
	if (tonumber(build) < 7897) then
		self:RemoveAllParserListeners()
		self:AddParserListener({ eventType = "Cast", sourceName_not = false, abilityName_not = false, isBegin = true }, "OnCastEvent")
		self:AddParserListener({ eventType = "Aura", recipientName_not = false, abilityName_not = false, isBuff = true }, "OnBuffEvent")
		self:AddParserListener({ eventType = "Aura", recipientName_not = false, abilityName_not = false, isBuff = false }, "OnDebuffEvent")
		self:AddParserListener({ eventType = "Dispel", recipientName_not = false, recipientAbilityName_not = false }, "OnDispelEvent")
		self:AddParserListener({ eventType = "Fade", recipientName_not = false, abilityName_not = false }, "OnFadeEvent")
	else
		self:AddEventListener("COMBAT_LOG_EVENT_UNFILTERED", "CombatLogEvent")
	end
	self:AddEventListener("PLAYER_DEAD", "UpdateUnitInfo")
	self:AddEventListener("UNIT_COMBAT", "SetUnitInfo")
	self:AddEventListener("UNIT_TARGET", "SetUnitInfo")
	self:AddEventListener("PLAYER_ENTERING_WORLD")
	self:AddEventListener("PLAYER_REGEN_ENABLED")
	self:AddEventListener("PLAYER_REGEN_DISABLED")
	self:AddEventListener("BAG_UPDATE_COOLDOWN")
	self:AddEventListener("SPELL_UPDATE_COOLDOWN")
	if self.db.profile.preset or self.db.profile.damage then
		self:AddHook("CombatText_AddMessage")
	end
	self.feedback.RegisterCallback(self.callbacks, "OnFeedbackFrameMoved")
end
